@@ -20,7 +20,7 @@ from pkg_resources import iter_entry_points
 from twarc.version import version
 from twarc.handshake import handshake
 from twarc.decorators import cli_api_error
-from twarc.expansions import flatten as flat
+from twarc.expansions import ensure_flattened
 from click_config_file import configuration_option
 
 
@@ -150,22 +150,20 @@ def get_version():
     help='Match tweets sent prior to tweet id')
 @click.option('--start-time',
     type=click.DateTime(formats=('%Y-%m-%d', '%Y-%m-%dT%H:%M:%S')),
-    help='Match tweets created after time (ISO 8601/RFC 3339), e.g.  2021-01-01T12:31:04')
+    help='Match tweets created after UTC time (ISO 8601/RFC 3339), e.g.  2021-01-01T12:31:04')
 @click.option('--end-time',
     type=click.DateTime(formats=('%Y-%m-%d', '%Y-%m-%dT%H:%M:%S')),
-    help='Match tweets sent before time (ISO 8601/RFC 3339)')
+    help='Match tweets sent before UTC time (ISO 8601/RFC 3339)')
 @click.option('--archive', is_flag=True, default=False,
     help='Search the full archive (requires Academic Research track)')
 @click.option('--limit', default=0, help='Maximum number of tweets to save')
 @click.option('--max-results', default=0, help='Maximum number of tweets per API response')
-@click.option('--flatten', is_flag=True, default=False,
-    help='Include expansions inline with tweets, and one line per tweet')
 @click.argument('query', type=str)
 @click.argument('outfile', type=click.File('w'), default='-')
 @click.pass_obj
 @cli_api_error
 def search(T, query, outfile, since_id, until_id, start_time, end_time, limit,
-        max_results, archive, flatten):
+        max_results, archive):
     """
     Search for tweets.
     """
@@ -177,12 +175,6 @@ def search(T, query, outfile, since_id, until_id, start_time, end_time, limit,
         # default number of tweets per response 500 when not set otherwise
         if max_results == 0:
             max_results = 500
-
-        # if the user is searching the historical archive the assumption is that
-        # they want to search everything, and not just the previous month which
-        # is the default: https://github.com/DocNow/twarc/issues/434
-        if start_time == None and since_id == None:
-            start_time = datetime.datetime(2006, 3, 21, tzinfo=datetime.timezone.utc)
     else:
         if max_results == 0:
             max_results = 100
@@ -190,21 +182,19 @@ def search(T, query, outfile, since_id, until_id, start_time, end_time, limit,
 
     for result in search_method(query, since_id, until_id, start_time, end_time,
             max_results):
-        _write(result, outfile, flatten)
+        _write(result, outfile)
         count += len(result['data'])
         if limit != 0 and count >= limit:
             break
 
 @twarc2.command('tweet')
-@click.option('--flatten', is_flag=True, default=False,
-    help='Include expansions inline with tweets, and one line per tweet')
 @click.option('--pretty', is_flag=True, default=False,
     help='Pretty print the JSON')
 @click.argument('tweet_id', type=str)
 @click.argument('outfile', type=click.File('w'), default='-')
 @click.pass_obj
 @cli_api_error
-def tweet(T, tweet_id, outfile, flatten, pretty):
+def tweet(T, tweet_id, outfile, pretty):
     """
     Look up a tweet using its tweet id or URL.
     """
@@ -213,25 +203,23 @@ def tweet(T, tweet_id, outfile, flatten, pretty):
     if not re.match('^\d+$', tweet_id):
         click.echo(click.style("Please enter a tweet URL or ID", fg="red"), err=True)
     result = next(T.tweet_lookup([tweet_id]))
-    _write(result, outfile, flatten, pretty=pretty)
+    _write(result, outfile, pretty=pretty)
 
 
 @twarc2.command('followers')
 @click.option('--limit', default=0, help='Maximum number of followers to save')
-@click.option('--flatten', is_flag=True, default=False,
-    help='Include expansions inline with users, and one line per user')
 @click.argument('user', type=str)
 @click.argument('outfile', type=click.File('w'), default='-')
 @click.pass_obj
 @cli_api_error
-def followers(T, user, outfile, limit, flatten):
+def followers(T, user, outfile, limit):
     """
     Get the followers for a given user.
     """
     count = 0
 
     for result in T.followers(user):
-        _write(result, outfile, flatten)
+        _write(result, outfile)
         count += len(result['data'])
         if limit != 0 and count >= limit:
             break
@@ -239,20 +227,18 @@ def followers(T, user, outfile, limit, flatten):
 
 @twarc2.command('following')
 @click.option('--limit', default=0, help='Maximum number of friends to save')
-@click.option('--flatten', is_flag=True, default=False,
-    help='Include expansions inline with users, and one line per user')
 @click.argument('userd', type=str)
 @click.argument('outfile', type=click.File('w'), default='-')
 @click.pass_obj
 @cli_api_error
-def following(T, user, outfile, limit, flatten):
+def following(T, user, outfile, limit):
     """
     Get the users who are following a given user.
     """
     count = 0
 
     for result in T.following(user):
-        _write(result, outfile, flatten)
+        _write(result, outfile)
         count += len(result['data'])
         if limit != 0 and count >= limit:
             break
@@ -260,12 +246,10 @@ def following(T, user, outfile, limit, flatten):
 
 @twarc2.command('sample')
 @click.option('--limit', default=0, help='Maximum number of tweets to save')
-@click.option('--flatten', is_flag=True, default=False,
-    help='Include expansions inline with tweets, and one line per tweet.')
 @click.argument('outfile', type=click.File('a+'), default='-')
 @click.pass_obj
 @cli_api_error
-def sample(T, flatten, outfile, limit):
+def sample(T, outfile, limit):
     """
     Fetch tweets from the sample stream.
     """
@@ -276,38 +260,35 @@ def sample(T, flatten, outfile, limit):
         count += 1
         if limit != 0 and count >= limit:
             event.set()
-        _write(result, outfile, flatten)
+        _write(result, outfile)
 
 
 @twarc2.command('hydrate')
 @click.argument('infile', type=click.File('r'), default='-')
 @click.argument('outfile', type=click.File('w'), default='-')
-@click.option('--flatten', is_flag=True, default=False,
-    help='Include expansions inline with tweets, and one line per tweet.')
 @click.pass_obj
 @cli_api_error
-def hydrate(T, infile, outfile, flatten):
+def hydrate(T, infile, outfile):
     """
     Hydrate tweet ids.
     """
     for result in T.tweet_lookup(infile):
-        _write(result, outfile, flatten)
+        _write(result, outfile)
 
 
 @twarc2.command('users')
 @click.option('--usernames', is_flag=True, default=False)
-@click.option('--flatten', is_flag=True, default=False,
-    help='Include expansions inline with tweets, and one line per tweet.')
 @click.argument('infile', type=click.File('r'), default='-')
 @click.argument('outfile', type=click.File('w'), default='-')
 @click.pass_obj
 @cli_api_error
-def users(T, infile, outfile, usernames, flatten):
+def users(T, infile, outfile, usernames):
     """
     Get data for user ids or usernames.
     """
     for result in T.user_lookup(infile, usernames):
-        _write(result, outfile, flatten)
+        _write(result, outfile)
+
 
 @twarc2.command('mentions')
 @click.option('--since-id', type=int,
@@ -320,20 +301,20 @@ def users(T, infile, outfile, usernames, flatten):
 @click.option('--end-time',
     type=click.DateTime(formats=('%Y-%m-%d', '%Y-%m-%dT%H:%M:%S')),
     help='Match tweets sent before time (ISO 8601/RFC 3339)')
-@click.option('--flatten', is_flag=True, default=False,
-    help='Include expansions inline with tweets, and one line per tweet')
 @click.argument('user_id', type=str)
 @click.argument('outfile', type=click.File('w'), default='-')
 @click.pass_obj
 @cli_api_error
-def mentions(T, user_id, outfile, since_id, until_id, start_time, end_time, flatten):
+def mentions(T, user_id, outfile, since_id, until_id, start_time, end_time):
     """
     Retrieve the most recent tweets mentioning the given user.
     """
     for result in T.mentions(user_id, since_id, until_id, start_time, end_time):
-        _write(result, outfile, flatten)
+        _write(result, outfile)
+
 
 @twarc2.command('timeline')
+@click.option('--limit', default=0, help='Maximum number of tweets to return')
 @click.option('--since-id', type=int,
     help='Match tweets sent after tweet id')
 @click.option('--until-id', type=int,
@@ -344,18 +325,179 @@ def mentions(T, user_id, outfile, since_id, until_id, start_time, end_time, flat
 @click.option('--end-time',
     type=click.DateTime(formats=('%Y-%m-%d', '%Y-%m-%dT%H:%M:%S')),
     help='Match tweets sent before time (ISO 8601/RFC 3339)')
-@click.option('--flatten', is_flag=True, default=False,
-    help='Include expansions inline with tweets, and one line per tweet')
+@click.option('--use-search', is_flag=True, default=False,
+    help='Use the search/all API endpoint which is not limited to the last 3200 tweets, but requires Academic Product Track access.')
 @click.argument('user_id', type=str)
 @click.argument('outfile', type=click.File('w'), default='-')
 @click.pass_obj
 @cli_api_error
-def timeline(T, user_id, outfile, since_id, until_id, start_time, end_time, flatten):
+def timeline(T, user_id, outfile, since_id, until_id, start_time, end_time,
+        use_search, limit):
     """
-    Retrieve the 3200 most recent tweets for the given user.
+    Retrieve st recent tweets for the given user.
     """
-    for result in T.timeline(user_id, since_id, until_id, start_time, end_time):
-        _write(result, outfile, flatten)
+
+    if use_search:
+        q = f'from:{user_id}'
+        tweets = T.search_all(q, since_id, until_id, start_time, end_time)
+    else:
+        tweets = T.timeline(user_id, since_id, until_id, start_time, end_time)
+
+    count = 0
+    for result in tweets:
+        _write(result, outfile)
+
+        count += len(result['data'])
+        if limit != 0 and count >= limit:
+            break
+
+
+@twarc2.command('timelines')
+@click.option('--limit', default=0, help='Maximum number of tweets to return')
+@click.option('--timeline-limit', default=0,
+    help='Maximum number of tweets to return per-timeline')
+@click.option('--use-search', is_flag=True, default=False,
+    help='Use the search/all API endpoint which is not limited to the last 3200 tweets, but requires Academic Product Track access.')
+@click.argument('infile', type=click.File('r'), default='-')
+@click.argument('outfile', type=click.File('w'), default='-')
+@click.pass_obj
+def timelines(T, infile, outfile, limit, timeline_limit, use_search):
+    """
+    Fetch the timelines of every user in an input source of tweets. If
+    the input is a line oriented text file of user ids or usernames that will 
+    be used instead.
+    """
+    total_count = 0
+    seen = set()
+    for line in infile:
+        line = line.strip()
+
+        users = []
+        try:
+            data = ensure_flattened(json.loads(line))
+            users = set([t['author']['id'] for t in ensure_flattened(data)])
+        except json.JSONDecodeError:
+            users = set([line])
+
+        for user in users:
+
+            # only process a given user once
+            if user in seen:
+                continue
+            seen.add(user)
+
+            # which api endpoint to use
+            if use_search and since_id:
+                tweets = T.search_all(f'from:{user}', since_id=since_id)
+            elif use_search:
+                tweets = T.search_all(f'from:{user}')
+            else:
+                tweets = T.timeline(user)
+
+            timeline_count = 0
+            for response in tweets:
+                _write(response, outfile)
+
+                timeline_count += len(response['data'])
+                if timeline_limit != 0 and timeline_count >= timeline_limit:
+                    break
+
+                total_count += len(response['data'])
+                if limit != 0 and total_count >= limit:
+                    return
+
+
+@twarc2.command('conversation')
+@click.option('--archive', is_flag=True, default=False,
+    help='Search the full archive (requires Academic Research track)')
+@click.argument('tweet_id', type=str)
+@click.argument('outfile', type=click.File('w'), default='-')
+@click.pass_obj
+@cli_api_error
+def conversation(T, tweet_id, archive, outfile):
+    """
+    Retrieve a conversation thread using the tweet id.
+    """
+    q = f'conversation_id:{tweet_id}'
+    if archive:
+        search = T.search_all(q)
+    else:
+        search = T.search_recent(q)
+    for resp in search:
+        _write(resp, outfile)
+
+
+@twarc2.command('conversations')
+@click.option('--limit', default=0, help='Maximum number of tweets to return')
+@click.option('--conversation-limit', default=0,
+    help='Maximum number of tweets to return per-conversation')
+@click.option('--archive', is_flag=True, default=False,
+    help='Use the Academic Research project track access to the full archive')
+@click.argument('infile', type=click.File('r'), default='-')
+@click.argument('outfile', type=click.File('w'), default='-')
+@click.pass_obj
+@cli_api_error
+def conversations(T, infile, outfile, archive, limit, conversation_limit):
+    """
+    Fetch the full conversation threads that the input tweets are a part of.
+    Alternatively the input can be a line oriented file of conversation ids.
+    """
+
+    # keep track of converstation ids that have been fetched so that they
+    # aren't fetched twice
+    seen = set()
+
+    # use the archive or recent search?
+    search = T.search_all if archive else T.search_recent
+
+    count = 0
+    stop = False
+    for line in infile:
+        conv_ids = []
+
+        # stop will get set when the total tweet limit has been met
+        if stop:
+            break
+
+        # get a specific conversation id
+        line = line.strip()
+        if re.match(r'^\d+$', line):
+            if line in seen:
+                continue
+            conv_ids = [line]
+
+        # generate all conversation_ids that are referenced in tweets input
+        else:
+            def f():
+                for tweet in ensure_flattened(json.loads(line)):
+                    yield tweet.get('conversation_id')
+            conv_ids = f()
+
+        # output results while paying attention to the set limits
+        conv_count = 0
+
+        for conv_id in conv_ids:
+
+            if conv_id in seen:
+                logging.info(f'already fetched conversation_id {conv_id}')
+            seen.add(conv_id)
+
+            conv_count = 0
+
+            logging.info(f'fetching conversation {conv_id}')
+            for result in search(f'conversation_id:{conv_id}'): 
+                _write(result, outfile, False)
+
+                count += len(result['data'])
+                if limit != 0 and count >= limit:
+                    logging.info(f'reached tweet limit of {limit}')
+                    stop = True
+                    break
+
+                conv_count += len(result['data'])
+                if conversation_limit !=0 and conv_count >= conversation_limit:
+                    logging.info(f'reached conversation limit {conversation_limit}')
+                    break
 
 
 @twarc2.command('flatten')
@@ -364,39 +506,40 @@ def timeline(T, user_id, outfile, since_id, until_id, start_time, end_time, flat
 @cli_api_error
 def flatten(infile, outfile):
     """
-    "Flatten" tweets, or move expansions inline with tweet objects.
+    "Flatten" tweets, or move expansions inline with tweet objects and ensure
+    that each line of output is a signle tweet.
     """
     if (infile.name == outfile.name):
         click.echo(click.style(f"💔 Cannot flatten files in-place, specify a different output file!", fg='red'), err=True)
         return
 
     for line in infile:
-        result = json.loads(line)
-        _write(result, outfile, True)
+        for tweet in ensure_flattened(json.loads(line)):
+            _write(tweet, outfile, False)
 
 
 @twarc2.command('stream')
 @click.option('--limit', default=0, help='Maximum number of tweets to return')
-@click.option('--flatten', is_flag=True, default=False,
-    help='Include expansions inline with tweets, and one line per tweet')
 @click.argument('outfile', type=click.File('a+'), default='-')
 @click.pass_obj
 @cli_api_error
-def stream(T, flatten, outfile, limit):
+def stream(T, outfile, limit):
     """
     Fetch tweets from the live stream.
     """
     event = threading.Event()
     count = 0
-    click.echo(click.style(f'Started a stream with rules:', fg='green'))
+    click.echo(click.style(f'Started a stream with rules:', fg='green'),
+            err=True)
     _print_stream_rules(T)
-    click.echo(click.style(f'Writing to {outfile.name}\nCTRL+C to stop...', fg='green'))
+    click.echo(click.style(f'Writing to {outfile.name}\nCTRL+C to stop...',
+        fg='green'), err=True)
     for result in T.stream(event=event):
         count += 1
         if limit != 0 and count == limit:
             logging.info(f'reached limit {limit}')
             event.set()
-        _write(result, outfile, flatten)
+        _write(result, outfile)
 
 
 @twarc2.group()
@@ -423,7 +566,7 @@ def _print_stream_rules(T):
     """
     result = T.get_stream_rules()
     if 'data' not in result or len(result['data']) == 0:
-        click.echo('No rules yet. Add them with ' + click.style('twarc2 stream-rules add', bold=True))
+        click.echo('No rules yet. Add them with ' + click.style('twarc2 stream-rules add', bold=True), err=True)
     else:
         count = 0
         for rule in result['data']:
@@ -432,7 +575,7 @@ def _print_stream_rules(T):
             s = rule['value']
             if 'tag' in rule:
                 s += f" (tag: {rule['tag']})"
-            click.echo(click.style(f'☑  {s}'))
+            click.echo(click.style(f'☑  {s}'), err=True)
             count += 1
 
 
@@ -534,17 +677,6 @@ def _error_str(errors):
 
     return click.style("\n".join(parts), fg="red")
 
-def _write(results, outfile, flatten, pretty=False):
+def _write(results, outfile, pretty=False):
     indent = 2 if pretty else None
-    if 'data' in results:
-        if flatten:
-            if isinstance(results['data'], list):
-                for r in flat(results)['data']:
-                    click.echo(json.dumps(r, indent=indent), file=outfile)
-            else:
-                r = flat(results)['data']
-                click.echo(json.dumps(r, indent=indent), file=outfile)
-        else:
-            click.echo(json.dumps(results, indent=indent), file=outfile)
-    else:
-        click.echo(json.dumps(results, indent=indent), file=outfile)
+    click.echo(json.dumps(results, indent=indent), file=outfile)
